//=================================================
//
// Terrain Generation Using Fluid Simulation
// by Francis "DeathWish" Woodhouse
// deathwish@valve-erc.com / http://www.dwish.net
//
// This is a sample application that implements the algorithm
// described in the article.
//
//=================================================

#include <stdio.h>
#include <stdlib.h>

float* g_pSurface[2] = {NULL,NULL}; // surface at time k and k-1

// basic parameters
int g_iSurfW=0, g_iSurfH=0, g_iNumIts=0;
float g_fMinInitH=-500, g_fMaxInitH=500;

// parameters to the fluid simulation equation
float d = 10.0f;
float t = 0.033f;
float mu = 100.0f;
float c = 100.0f;

// pre-computed coefficients in the fluid equation
float coefA, coefB, coefC;

int main(int argc, char* argv[])
{
	// check that they've got the right number of arguments
	if (argc < 5) {
		printf("Usage:\n");
		printf("LiquidTerrainGen outputfile width height iterations [seed] [min initial height] [max initial height] [viscosity] [wave speed] [timestep] [node distance]\n\n");
		exit(0);
	}

	// set the mandatory argument variables' values
	g_iSurfW = atoi(argv[2]);
	g_iSurfH = atoi(argv[3]);
	g_iNumIts = atoi(argv[4]);

	// edit the value of the optional arguments if they're specified
	if (argc >= 6)
		srand(atoi(argv[5]));
	if (argc >= 7)
		g_fMinInitH = atof(argv[6]);
	if (argc >= 8)
		g_fMaxInitH = atof(argv[7]);
	if (argc >= 9)
		mu = atof(argv[8]);
	if (argc >= 10)
		c = atof(argv[9]);
	if (argc >= 11)
		t = atof(argv[10]);
	if (argc >= 12)
		d = atof(argv[11]);

	// allocate memory for the surface at two timesteps
	g_pSurface[0] = new float[g_iSurfW*g_iSurfH];
	g_pSurface[1] = new float[g_iSurfW*g_iSurfH];

	// pre-compute the coefficients in the fluid equation
	coefA = (4 - (8*c*c*t*t) / (d*d)) / (mu*t + 2);
	coefB = (mu*t - 2) / (mu*t + 2);
	coefC = ((2*c*c*t*t) / (d*d)) / (mu*t + 2);

	// initialise the surfaces to random values, except the edges which are set to 0
	for (int y = 0; y < g_iSurfH; y++) {
		for (int x = 0; x < g_iSurfW; x++) {
			if (x == 0 || x == (g_iSurfW-1) || y == 0 || y == (g_iSurfH-1))
				g_pSurface[0][x+y*g_iSurfW] = g_pSurface[1][x+y*g_iSurfW] = 0;
			else
				g_pSurface[0][x+y*g_iSurfW] = g_pSurface[1][x+y*g_iSurfW] = ((float)rand() / RAND_MAX) * (g_fMaxInitH - g_fMinInitH) + g_fMinInitH;
		}
	}

	int iCurBuf = 0;
	int iInd;

	float* pOld, * pNew;

	// Iterate over the heightmap, applying the fluid simulation equation.
	// Although it requires knowledge of the two previous timesteps, it only
	// accesses one pixel of the k-1 timestep, so using a simple trick we only
	// need to store the heightmap twice, not three times, and we can avoid
	// a memcpy() every iteration.
	for (int i = 0; i < g_iNumIts; i++) {
		pOld = g_pSurface[1-iCurBuf];
		pNew = g_pSurface[iCurBuf];

		for (int y = 1; y < g_iSurfH-1; y++) {
			for (int x = 1; x < g_iSurfW-1; x++) {
				iInd = x+y*g_iSurfW;

				pOld[iInd] = coefA*pNew[iInd] + coefB*pOld[iInd] +
							 coefC*(pNew[iInd+1] + pNew[iInd-1] + pNew[iInd+g_iSurfW] + pNew[iInd-g_iSurfW]);
			}
		}

		iCurBuf = 1-iCurBuf;
	}

	float fMinH=g_pSurface[iCurBuf][0], fMaxH=g_pSurface[iCurBuf][0];

	// find the minimum and maximum heights
	for (y = 0; y < g_iSurfH; y++) {
		for (int x = 0; x < g_iSurfW; x++) {
			if (g_pSurface[iCurBuf][x+y*g_iSurfW] > fMaxH)
				fMaxH = g_pSurface[iCurBuf][x+y*g_iSurfW];
			else if (g_pSurface[iCurBuf][x+y*g_iSurfW] < fMinH)
				fMinH = g_pSurface[iCurBuf][x+y*g_iSurfW];
		}
	}

	// normalize the surface
	for (y = 0; y < g_iSurfH; y++) {
		for (int x = 0; x < g_iSurfW; x++)
			g_pSurface[iCurBuf][x+y*g_iSurfW] = (g_pSurface[iCurBuf][x+y*g_iSurfW] - fMinH) / (fMaxH - fMinH);
	}

	// allocate memory for the greyscale image
	unsigned char* pImg = new unsigned char[g_iSurfW*g_iSurfH];

	// put the normalized heightmap into the range [0...255] and into the greyscale image
	for (y = 0; y < g_iSurfH; y++) {
		for (int x = 0; x < g_iSurfW; x++)
			pImg[x+y*g_iSurfW] = (unsigned char)(g_pSurface[iCurBuf][x+y*g_iSurfW] * 255);
	}

	// output the image as raw data
	FILE* pFile = fopen(argv[1], "wb");
	if (pFile) {
		fwrite(pImg, 1, g_iSurfW*g_iSurfH, pFile);
		fclose(pFile);

		printf("Success!\n\n");
	}
	else
		printf("ERROR: Could not open %s for binary write!\n\n", argv[1]);

	// deallocate memory
	delete[] pImg;
	delete[] g_pSurface[0];
	delete[] g_pSurface[1];

	return 0;
}